unit VirtualKeyboard;

{$mode objfpc}{$H+}

interface

uses
    Classes
  , SysUtils
  , LResources
  , Controls
  , Buttons
  , LCLType
  ;

const
  KEY_DEFAULT_WIDTH = 1;

type
  { TOwnedCollection }

  TOwnedCollection = class(TCollection)
  strict private
    fOwner: TPersistent;
  protected
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  private // may be set only by members of this unit
    procedure SetOwner(aOwner: TPersistent);
  end;

  { TKeyAlternative }

  TKeyAlternative = class(TCollectionItem)
  strict private
    fCharacter: TUTF8Char;
    fDisplayStr: UTF8String;
    fShift: TShiftState;
    procedure SetCaption(AValue: UTF8String);
    procedure SetCharacter(AValue: TUTF8Char);
    procedure SetShift(AValue: TShiftState);
  published
    // the TShiftState when this alternative should be shown
    property Shift: TShiftState read fShift write SetShift default [];
    // the character it represents
    property Character: TUTF8Char read fCharacter write SetCharacter;
    // Caption to display; if not set, @Character will be uses
    property Caption: UTF8String read fDisplayStr write SetCaption;
  end;

type
  { TKeyboardKey }

  TKeyboardKey = class(TCollectionItem, IFPObserver)
  strict private
    fAlternatives: TOwnedCollection;
    // Width = 0 -> don't show
    // default: KEY_DEFAULT_WIDTH
    fButtonWidth: Byte;
    // quick access to the specified alternative
    fCurrentShiftState: TShiftState;
    fCurrentAlternative: TKeyAlternative;
    function GetCurrentChar: TUTF8Char; inline;
    function GetCurrentDisplayStr: UTF8String; inline;
    procedure SetAlternatives(AValue: TOwnedCollection);
    procedure SetButtonWidthProp(AValue: Byte);
    procedure SetCurrentShiftState(AValue: TShiftState);
  public
    constructor Create(ACollection: TCollection); override;
    destructor Destroy; override;
    // returns the instance for chaining
    function Add(const aChar: TUTF8Char; const aDisplayStr: UTF8String;
      aShift: TShiftState): TKeyboardKey;
    function Add(const aChar: TUTF8Char; aShift: TShiftState): TKeyboardKey;
    function Add(const aChar: TUTF8Char): TKeyboardKey;
    function AddShift(const aChar: TUTF8Char): TKeyboardKey;
    procedure FPOObservedChanged(ASender: TObject;
      Operation: TFPObservedOperation; Data: Pointer);
    function SetButtonWidth(aWidth: Byte): TKeyboardKey;
    property CurrentShiftState: TShiftState read fCurrentShiftState write SetCurrentShiftState;
    property CurrentAlternative: TKeyAlternative read fCurrentAlternative;
    property CurrentDisplayStr: UTF8String read GetCurrentDisplayStr;
    property CurrentChar: TUTF8Char read GetCurrentChar;
  public
    class function NewKey(aColl: TCollection): TKeyboardKey;
    class function NewKey(aColl: TCollection; const aChar: TUTF8Char): TKeyboardKey;
  published
    property Alternatives: TOwnedCollection read fAlternatives write SetAlternatives;
    property ButtonWidth: Byte read fButtonWidth write SetButtonWidthProp;
  end;

type

  { TLayoutLine }

  TLayoutLine = class(TCollectionItem, IFPObserver)
  strict private
    fKeys: TOwnedCollection;
    procedure SetKeys(AValue: TOwnedCollection);
  public
    constructor Create(ACollection: TCollection); override;
    destructor Destroy; override;
    procedure FPOObservedChanged(ASender: TObject;
      Operation: TFPObservedOperation; Data: Pointer);
    function NewKey: TKeyboardKey;
    function NewKey(const aChar: TUTF8Char): TKeyboardKey;
  published
    property Keys: TOwnedCollection read fKeys write SetKeys;
  end;

  TKeyboardLayout = TOwnedCollection;

  { TVirtualKeyboard }

  TVirtualKeyboard = class(TCustomControl, IFPObserver)
  strict private type
    TKeyboardButton = record
      Button: TSpeedButton;
      Key: TKeyboardKey;
    end;

    TKeyLine = array of TKeyboardButton;
    TKeyMap = array of TKeyLine;

  strict private
    // external defined keys
    fKeyLayout: TKeyboardLayout;
    // internal mapping of controls to keys
    fKeys: TKeyMap;
    // current shift state
    fShift: TShiftState;
    procedure DoOnKeyPress(Sender: TObject);
    procedure SetKeyLayout(AValue: TKeyboardLayout);
    procedure SetShiftState(AValue: TShiftState);
  strict protected
    procedure RemoveKeyControls; virtual;
    procedure DoInternalKeyHandling(const aKey: TUTF8Char); virtual;
    function GetTextHeight: Integer; virtual;
  protected
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    procedure Loaded; override;
    procedure CreateWnd; override;
    procedure DoAutoSize; override;
    function GetButtonHeight: Integer; virtual;
    function GetButtonWidth(aFactor: Byte): Integer;
  public
    procedure FPOObservedChanged(ASender: TObject;
      Operation: TFPObservedOperation; Data: Pointer);
    // call after direct modifications of @KeyLayout
    procedure RebuildKeys; virtual;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CalculatePreferredSize(var PreferredWidth,
      PreferredHeight: integer; WithThemeSpace: Boolean); override;
  published
    property Anchors;
    property AutoSize;
    property BorderSpacing;
    property OnUTF8KeyPress;
    property OnKeyPress;
    property ShiftState: TShiftState read fShift write SetShiftState;
    property Font;
    property KeyLayout: TKeyboardLayout read fKeyLayout write SetKeyLayout;
    property ParentFont;
  end;

  TDefineLayout = procedure(Layout: TKeyboardLayout);

procedure Register;

var
  OnDefaultLayout: TDefineLayout;

implementation

uses
    LazUTF8
  ;

procedure Register;
begin
  {$I virtualkeyboard_icon.lrs}
  RegisterComponents('Touchhelper',[TVirtualKeyboard]);
end;

function TOwnedCollection.GetOwner: TPersistent;
begin
  Result := fOwner;
end;

procedure TOwnedCollection.Update(Item: TCollectionItem);
begin
  if UpdateCount = 0 then
    inherited Update(Item);
end;

procedure TOwnedCollection.SetOwner(aOwner: TPersistent);
begin
  fOwner := aOwner;
end;

{ TKeyAlternative }

procedure TKeyAlternative.SetCaption(AValue: UTF8String);
begin
  if fDisplayStr = AValue then Exit;
  fDisplayStr := AValue;
  Changed(False);
end;

procedure TKeyAlternative.SetCharacter(AValue: TUTF8Char);
begin
  if fCharacter = AValue then Exit;
  fCharacter := AValue;
  Changed(False);
end;

procedure TKeyAlternative.SetShift(AValue: TShiftState);
begin
  if fShift = AValue then Exit;
  fShift := AValue;
  Changed(False);
end;

{ TLayoutLine }

procedure TLayoutLine.SetKeys(AValue: TOwnedCollection);
begin
  if fKeys = AValue then Exit;
  fKeys.Free;
  fKeys := AValue;
  fKeys.SetOwner(Self);
  Changed(True);
end;

constructor TLayoutLine.Create(ACollection: TCollection);
begin
  inherited Create(nil);
  fKeys := TOwnedCollection.Create(TKeyboardKey);
  fKeys.FPOAttachObserver(Self);
  fKeys.SetOwner(Self);
  // delay addition to the collection after the creation of all child objects
  // the addition triggers a change in the collection which will be propageted
  // to the keyboard which tries to rebuild and accesses the @fKeys field which
  // isn't initialized at this point
  Self.Collection := ACollection;
end;

destructor TLayoutLine.Destroy;
begin
  fKeys.Free;
  inherited Destroy;
end;

procedure TLayoutLine.FPOObservedChanged(ASender: TObject;
  Operation: TFPObservedOperation; Data: Pointer);
begin
  Assert(ASender is TKeyboardKey, 'Sender not expected.');
  if UpdateCount = 0 then
    Changed(False);
end;

function TLayoutLine.NewKey: TKeyboardKey;
begin
  Result := TKeyboardKey(fKeys.Add);
end;

function TLayoutLine.NewKey(const aChar: TUTF8Char): TKeyboardKey;
begin
  Result := NewKey().Add(aChar);
end;

{ TKeyboardKey }

function TKeyboardKey.GetCurrentChar: TUTF8Char;
begin
  if Assigned(fCurrentAlternative) then
    Result := fCurrentAlternative.Character
  else
    Result := '';
end;

function TKeyboardKey.GetCurrentDisplayStr: UTF8String;
begin
  if Assigned(fCurrentAlternative) then
    Result := fCurrentAlternative.Caption
  else
    Result := '';
end;

procedure TKeyboardKey.SetAlternatives(AValue: TOwnedCollection);
begin
  if fAlternatives = AValue then Exit;
  fAlternatives.Free;
  fAlternatives := AValue;
  fAlternatives.SetOwner(Self);
  Changed(True);
end;

procedure TKeyboardKey.SetButtonWidthProp(AValue: Byte);
begin
  if fButtonWidth = AValue then Exit;
  fButtonWidth := AValue;
  Changed(False);
end;

procedure TKeyboardKey.SetCurrentShiftState(AValue: TShiftState);
var
  i: Integer;
begin
  if fCurrentShiftState = AValue then Exit;
  fCurrentShiftState := AValue;

  // use default if no match found
  if fAlternatives.Count > 0 then
    fCurrentAlternative := TKeyAlternative(fAlternatives.Items[0])
  else
    fCurrentAlternative := nil;

  for i := 0 to fAlternatives.Count - 1 do
    if TKeyAlternative(fAlternatives.Items[i]).Shift = AValue then
    begin
      fCurrentAlternative := TKeyAlternative(fAlternatives.Items[i]);
      break;
    end else
    if (aValue * TKeyAlternative(fAlternatives.Items[i]).Shift) <> [] then
      fCurrentAlternative := TKeyAlternative(fAlternatives.Items[i]);
end;

constructor TKeyboardKey.Create(ACollection: TCollection);
begin
  inherited Create(ACollection);
  fButtonWidth := 1;
  fAlternatives := TOwnedCollection.Create(TKeyAlternative);
  fAlternatives.FPOAttachObserver(Self);
  if Self.Collection is TOwnedCollection then
    fAlternatives.SetOwner(TOwnedCollection(Self.Collection).GetOwner)
  else
    fAlternatives.SetOwner(Self);
  fCurrentAlternative := nil;
end;

destructor TKeyboardKey.Destroy;
begin
  fAlternatives.Free;
  fCurrentAlternative := nil;
  inherited Destroy;
end;

function TKeyboardKey.Add(const aChar: TUTF8Char;
  const aDisplayStr: UTF8String; aShift: TShiftState): TKeyboardKey;
var
  theAlternative: TKeyAlternative;
begin
  theAlternative := TKeyAlternative.Create(fAlternatives);
  theAlternative.Character := aChar;
  theAlternative.Caption := aDisplayStr;
  theAlternative.Shift := aShift;

  Result := Self;
end;

function TKeyboardKey.Add(const aChar: TUTF8Char; aShift: TShiftState
  ): TKeyboardKey;
begin
  Result := Add(aChar, '', aShift);
end;

function TKeyboardKey.Add(const aChar: TUTF8Char): TKeyboardKey;
begin
  Result := Add(aChar, []);
end;

function TKeyboardKey.AddShift(const aChar: TUTF8Char): TKeyboardKey;
begin
  Result := Add(aChar, '', [ssShift]);
end;

procedure TKeyboardKey.FPOObservedChanged(ASender: TObject;
  Operation: TFPObservedOperation; Data: Pointer);
begin
  Assert(ASender is TKeyAlternative, 'Sender not expected.');

  if (fAlternatives.Count = 1) then
    fCurrentAlternative := TKeyAlternative(fAlternatives.Items[0]);

  Changed(False);
end;

function TKeyboardKey.SetButtonWidth(aWidth: Byte): TKeyboardKey;
begin
  fButtonWidth := aWidth;
  Result := Self;
end;

class function TKeyboardKey.NewKey(aColl: TCollection): TKeyboardKey;
begin
  Result := TKeyboardKey.Create(aColl);
end;

class function TKeyboardKey.NewKey(aColl: TCollection; const aChar: TUTF8Char
  ): TKeyboardKey;
begin
  Result := TKeyboardKey.Create(aColl);
  Result.Add(aChar, []);
end;

{ TVirtualKeyboard }

procedure TVirtualKeyboard.DoOnKeyPress(Sender: TObject);
var
  u8key: TUTF8Char;
  c: Char;
  KeyAlt: TKeyAlternative;
begin
  KeyAlt := TKeyAlternative(TSpeedButton(Sender).Tag);
  u8key := KeyAlt.Character;

  DoInternalKeyHandling(u8Key);
  if FindInvalidUTF8Character(@u8Key[1], length(u8Key), True) <> -1 then
  begin
    case ord(u8key[1]) of
      VK_LSHIFT, VK_RSHIFT: u8key := char(VK_SHIFT);
      VK_LCONTROL, VK_RCONTROL: u8Key := char(VK_CONTROL);
      VK_LMENU, VK_RMENU: u8Key := char(VK_MENU);
    end;
  end;
  DoUTF8KeyPress(u8Key);

  if length(u8key) = 0 then exit;

  // only for ASCII chars
  c := u8key[1];
  if ord(c) < $7F then
    KeyPress(c);
end;

procedure TVirtualKeyboard.SetKeyLayout(AValue: TKeyboardLayout);
begin
  if fKeyLayout = AValue then Exit;
  fKeyLayout.Free;
  fKeyLayout := AValue;
  RebuildKeys;
end;

procedure TVirtualKeyboard.SetShiftState(AValue: TShiftState);
begin
  if fShift = AValue then Exit;
  fShift := AValue;
end;

procedure TVirtualKeyboard.RebuildKeys;
var
  RowIdx: Integer;
  BtnIdx: Integer;
  Btn: TSpeedButton;
  PreviousButtonInRow: TSpeedButton;
  FirstButtonInLastRow: TSpeedButton;
  Row: TLayoutLine;
  key: TKeyboardKey;
begin
  RemoveKeyControls;
  InvalidatePreferredSize;

  if fKeyLayout.Count = 0 then exit;
  if csLoading in ComponentState then exit;
  if csDestroying in ComponentState then exit;

  SetLength(fKeys, fKeyLayout.Count);
  FillByte(fKeys[0], Length(fKeys) * Sizeof(fKeys[0]), 0);

  FirstButtonInLastRow := nil;
  for RowIdx := 0 to fKeyLayout.Count - 1 do
  begin
    Row := TLayoutLine(fKeyLayout.Items[RowIdx]);
    PreviousButtonInRow := nil;

    SetLength(fKeys[RowIdx], Row.Keys.Count);

    for BtnIdx := 0 to Row.Keys.Count - 1 do
    begin
      Btn := TSpeedButton.Create(Self);
      Btn.OnClick := @DoOnKeyPress;
      key := TKeyboardKey(Row.Keys.Items[BtnIdx]);
      key.CurrentShiftState := fShift;

      // cast to RawByteString because LCLs TCaption is still defined as
      // CP_ACP but expects UTF-8
      if key.CurrentDisplayStr <> '' then
        Btn.Caption := RawByteString(key.CurrentDisplayStr)
      else
        Btn.Caption := RawByteString(key.CurrentChar);

      Btn.Parent := Self;
      Btn.ParentFont := True;
      Btn.Height := GetButtonHeight;
      Btn.Width  := GetButtonWidth(key.ButtonWidth);
      Btn.Constraints.MinWidth := Btn.Width;
      Btn.Tag := PtrInt(key.CurrentAlternative);

      if Assigned(PreviousButtonInRow) then
      begin
        if BtnIdx = Row.Keys.Count - 1 then  // last button in Row
        begin
          Btn.Anchors := Btn.Anchors + [akRight];
          Btn.AnchorSide[akRight].Control := Self;
          Btn.AnchorSide[akRight].Side    := asrRight;
        end;
        // all buttons in row but not the first
        Btn.AnchorSide[akTop ].Control := PreviousButtonInRow;
        Btn.AnchorSide[akLeft].Side    := asrRight;
        Btn.AnchorSide[akLeft].Control := PreviousButtonInRow;
      end
      else
      begin
        if Assigned(FirstButtonInLastRow) then
        begin
          Btn.AnchorSide[akTop].Control := FirstButtonInLastRow;
          Btn.AnchorSide[akTop].Side    := asrBottom;
          if (RowIdx mod 2) = 1 then
          begin
            Btn.AnchorSide[akLeft].Control := FirstButtonInLastRow;
            //Btn.BorderSpacing.Left := FirstButtonInLastRow.Width div 2;
            Btn.Width := round(Btn.Width * 1.5);
          end else
            Btn.AnchorSide[akLeft].Control := Self;
        end
        else  // top left button
        begin
          Btn.AnchorSide[akTop ].Control := Self;
          Btn.AnchorSide[akLeft].Control := Self;
        end;
        FirstButtonInLastRow := Btn;
      end;

      PreviousButtonInRow := Btn;

      fKeys[RowIdx][BtnIdx].Button := Btn;
      fKeys[RowIdx][BtnIdx].Key    := key;
    end;
  end;
end;

procedure TVirtualKeyboard.RemoveKeyControls;
var
  i: Integer;
begin
  // clear controls
  SetLength(fKeys, 0);
  for i := Self.ComponentCount - 1 downto 0 do
    Self.Components[i].Destroy;
end;

procedure TVirtualKeyboard.DoInternalKeyHandling(const aKey: TUTF8Char);
  procedure ToggleFlag(aFlag: TShiftStateEnum);
  begin
   Integer(fShift) := Integer(fShift) xor Integer([aFlag]);
  end;
var
  key_code: Integer;
  oldShift: TShiftState;
begin
  oldShift := fShift;
  key_code := ord(aKey[1]);
  case key_code of
    VK_CAPITAL:
      begin
        ToggleFlag(ssCaps);
        if ssCaps in fShift then
          include(fShift, ssShift)
        else
          exclude(fShift, ssShift);
      end;
    VK_SHIFT, VK_LSHIFT, VK_RSHIFT:
      begin
        ToggleFlag(ssShift);
        if not (ssShift in fShift) then
          exclude(fShift, ssCaps);
      end;
    VK_CONTROL, VK_LCONTROL, VK_RCONTROL: ToggleFlag(ssCtrl);
    VK_MENU, VK_LMENU: ToggleFlag(ssAlt);
    VK_RMENU: ToggleFlag(ssAltGr);
  else
    if (ssShift in fShift) and not(ssCaps in fShift) then exclude(fShift, ssShift);
  end;

  if oldshift <> fShift then
    RebuildKeys;
end;

function TVirtualKeyboard.GetTextHeight: Integer;
const
  DEFAULT_HEIGHT = 12;
begin
  Result := Font.Height;
  if (Result = 0) and not Canvas.HandleAllocated then
    Result := DEFAULT_HEIGHT
  else
    Result := Canvas.TextHeight('X');
end;

procedure TVirtualKeyboard.GetChildren(Proc: TGetChildProc; Root: TComponent);
begin
  // do nothing; Buttons may not be streamed
end;

procedure TVirtualKeyboard.Loaded;
begin
  inherited Loaded;
  RebuildKeys;
end;

procedure TVirtualKeyboard.CreateWnd;
begin
  inherited CreateWnd;
  RebuildKeys;
end;

procedure TVirtualKeyboard.DoAutoSize;
begin
  inherited DoAutoSize;
end;

function TVirtualKeyboard.GetButtonHeight: Integer;
begin
  Result := GetTextHeight + 20;
end;

function TVirtualKeyboard.GetButtonWidth(aFactor: Byte): Integer;
begin
  Result := GetButtonHeight * aFactor;
end;

procedure TVirtualKeyboard.FPOObservedChanged(ASender: TObject;
  Operation: TFPObservedOperation; Data: Pointer);
begin
  Assert(ASender is TKeyboardLayout, 'Sender not expected.');
  if ([csLoading, csDestroying] * ComponentState) = [] then
    RebuildKeys;
end;

constructor TVirtualKeyboard.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fShift := [];
  fKeyLayout := TKeyboardLayout.Create(TLayoutLine);
  fKeyLayout.FPOAttachObserver(Self);
  fKeyLayout.SetOwner(Self);

  fKeyLayout.BeginUpdate;
  if Assigned(OnDefaultLayout) then
    OnDefaultLayout(fKeyLayout);
  fKeyLayout.EndUpdate;
end;

destructor TVirtualKeyboard.Destroy;
begin
  fKeyLayout.Free;
  inherited Destroy;
end;

procedure TVirtualKeyboard.CalculatePreferredSize(var PreferredWidth,
  PreferredHeight: integer; WithThemeSpace: Boolean);
var
  l: TCollectionItem;
  k: TCollectionItem;
  LWidth: Integer;
  MaxLWidth: Integer;
begin
  PreferredHeight := KeyLayout.Count * GetButtonHeight;
  MaxLWidth := 0;
  for l in KeyLayout do
  begin
    LWidth := 0;
    for k in TLayoutLine(l).Keys do
      LWidth := LWidth + GetButtonWidth(TKeyboardKey(k).ButtonWidth);
    if LWidth > MaxLWidth then
      MaxLWidth := LWidth;
  end;
  PreferredWidth := MaxLWidth;
end;

end.
