unit Numpad;

{$mode objfpc}{$H+}

interface

uses
    Classes
  , SysUtils
  , LResources
  , Forms
  , Controls
  , Graphics
  , Dialogs
  , Buttons
  ;

type
  TUndoIconSize = (
    uisNone,
    uis24,
    uis32,
    uis48,
    uis64,
    uis128,
    uis256,
    uis512
  );

  TDigit = 0..9;

  { Touch usable numeric input control }

  { TNumpad }

  TNumpad = class(TWinControl)
  strict private
    fBuffer: string;
    fOnChange: TNotifyEvent;
    fOnClear: TNotifyEvent;
    fUndoButton: TBitBtn;
    fUndoIconSize: TUndoIconSize;
    procedure SetUndoIconSize(AValue: TUndoIconSize);
    { Called by each button click }
    procedure ChildButtonClick(Sender: TObject);
  strict protected
    { Creates contained controls }
    procedure CreateSubControls; virtual;
    { Adds a decimal separator to the buffer

      @seealso(Buffer)
      @link(AddDigitToBuffer)
      @link(ChildButtonClick)
    }
    procedure AddDecimalSeparatorToBuffer; virtual;
    { Adds a digit to buffer

      Appends a digit to the end of the buffer. If the current @link(Buffer) is
      @code('0'), it will be replaced by any non-zero number.

      This method may be overriden in descendant classed for appropriate
      handling.

      @seealso(Buffer)
      @seealso(AddDecimalSeparatorToBuffer)
      @link(ChildButtonClick)
    }
    procedure AddDigitToBuffer(aDigit: TDigit); virtual;
    { Returns the default value

      The default value is an empty buffer @code('') but may be overridden in
      descendant classed.

      @seealso(Clear)
      @seealso(Buffer)
    }
    function GetClearedBuffer: String; virtual;
    { Updates Buffer

      Updates the buffer and calls the @link(OnChange) event if appliciable.
      @seealso(Buffer)
      @seealso(OnChange)
    }
    procedure SetBuffer(AValue: string); virtual;
  strict protected const
    { Tag property of reset button }
    TAG_RESET = 10;
    { Tag property of decimal separater button }
    TAG_DECIMALSEPARATOR = 11;
  public
    constructor Create(TheOwner: TComponent); override;
    { Assigns the default value

      The default value is determined by @link(GetClearedBuffer).

      @seealso(GetClearedBuffer)
    }
    procedure Clear; virtual;
    procedure InvalidatePreferredSize; override;
  published
    property Align;
    property Anchors;
    property BorderSpacing;
    // Buffer of digits already typed in
    property Buffer: string read fBuffer write SetBuffer;
    property Font;
    property ShowHint;
    { Size of the reset icon

      @seealso(TAG_RESET)
    }
    property UndoIconSize: TUndoIconSize read fUndoIconSize
      write SetUndoIconSize;
    property Visible;
    // Events
    { on each change

      Called by each change of the buffer independent if it's set by user or
      by code.

      @seealso(SetBuffer)
      @seealso(Buffer)
      @seealso(OnClick)
    }
    property OnChange: TNotifyEvent read fOnChange write fOnChange;
    // on reset button click
    property OnClear: TNotifyEvent read fOnClear write fOnClear;
    { on each button click

      Called on each button click independent if Buffer changed or not.

      For example: a click on reset without prior buffer change will call
      @link(OnClick) but not @link(OnChange)

      @seealso(OnChange)
    }
    property OnClick;
    property OnResize;
  end;

procedure Register;

implementation

{$R undo_icons.res}

uses
    ExtCtrls
  , Menus
  , resource
  , Math
  ;

var
  UndoImages: array[TUndoIconSize] of TPortableNetworkGraphic;

procedure Register;
begin
  {$I numpad_icon.lrs}
  RegisterComponents('Touchhelper', [TNumpad]);
end;

function UndoIconSizeToIconWidth(aSize: TUndoIconSize): integer; inline;
begin
  case aSize of
    uis24: Result := 24;
    uis32: Result := 32;
    uis48: Result := 48;
    uis64: Result := 64;
    uis128: Result := 128;
    uis256: Result := 256;
    uis512: Result := 512;
  else
    Result := 0;
  end;
end;

function GetNextUndoIconSize(aWidth, aHeight: integer): TUndoIconSize;
var
  s: TUndoIconSize;
  PixelSize: integer;
  BestDiff: int64;
  DeterminantSize: integer;
  DiffSize: int64;
begin
  BestDiff := High(BestDiff);
  DeterminantSize := min(aWidth, aHeight);
  Result := uis24;  // defalt value

  for s in TUndoIconSize do
  begin
    PixelSize := UndoIconSizeToIconWidth(s);
    DiffSize := abs(DeterminantSize - PixelSize);

    if (DiffSize = 0) then
      exit(s);

    if (DiffSize < BestDiff) and (PixelSize < DeterminantSize) then
    begin
      BestDiff := DiffSize;
      Result := s;
    end;
  end;
end;

function UndoIconSizeToResName(aSize: TUndoIconSize): string; inline;
var
  d: integer;
begin
  d := UndoIconSizeToIconWidth(aSize);
  Result := Format('EDIT_UNDO_%d', [d]);
end;

procedure InitUndoImages;
var
  i: TUndoIconSize;
  ResourceName: string;
begin
  UndoImages[uisNone] := nil;

  for i := succ(uisNone) to high(UndoImages) do
  begin
    ResourceName := UndoIconSizeToResName(i);
    UndoImages[i] := TPortableNetworkGraphic.Create;
    UndoImages[i].LoadFromResourceName(HINSTANCE, ResourceName);
  end;
end;

procedure FinalizeUndoImages;
var
  i: TUndoIconSize;
begin
  for i := low(UndoImages) to high(UndoImages) do
  begin
    UndoImages[i].Free;
    UndoImages[i] := nil;
  end;
end;

{ TNumpad }

procedure TNumpad.SetBuffer(AValue: string);
begin
  if fBuffer = AValue then
    Exit;
  fBuffer := AValue;
  if Assigned(fOnChange) then
    fOnChange(Self);
end;

procedure TNumpad.SetUndoIconSize(AValue: TUndoIconSize);
var
  png: TPortableNetworkGraphic;
begin
  if fUndoIconSize = AValue then Exit;
  fUndoIconSize := AValue;
  png := UndoImages[fUndoIconSize];
  fUndoButton.Glyph.Assign(png);
end;

procedure TNumpad.CreateSubControls;
var
  num: integer;
  Btn: TBitBtn;
begin
  // Numbers
  for num := 9 downto 0 do
  begin
    Btn := TBitBtn.Create(Self);
    Btn.Parent := Self;
    Btn.OnClick := @ChildButtonClick;

    if (num <> 0) and ((num mod 3) = 0) then
    begin
      Btn.Caption := IntToStr(num - 2);
      Btn.Tag := num - 2;
    end
    else if ((num + 2) mod 3) = 0 then
    begin
      Btn.Caption := IntToStr(num + 2);
      Btn.Tag := num + 2;
    end
    else
    begin
      Btn.Caption := IntToStr(num);
      Btn.Tag := num;
    end;
  end;

  // Back button
  fUndoButton := TBitBtn.Create(Self);
  fUndoButton.Parent := Self;
  fUndoButton.GlyphShowMode := gsmAlways;
  fUndoButton.Layout := blGlyphTop;
  fUndoButton.Tag := TAG_RESET;
  fUndoButton.OnClick := @ChildButtonClick;

  // Decimal Separator
  Btn := TBitBtn.Create(Self);
  Btn.Parent := Self;
  Btn.Caption := DefaultFormatSettings.DecimalSeparator;
  Btn.Tag := TAG_DECIMALSEPARATOR;
  Btn.OnClick := @ChildButtonClick;
end;

procedure TNumpad.ChildButtonClick(Sender: TObject);
var
  btn: TBitBtn absolute Sender;
begin
  Assert(Sender is TBitBtn, 'Expected subcontrols (TBitBtn).');
  case btn.Tag of
    TAG_RESET: Clear;
    TAG_DECIMALSEPARATOR: AddDecimalSeparatorToBuffer;
  else
    AddDigitToBuffer(btn.Tag);
  end;

  Click; // Trigger OnClick event
end;

procedure TNumpad.AddDecimalSeparatorToBuffer;
var
   p: SizeInt;
begin
  p := Pos(DefaultFormatSettings.DecimalSeparator, Buffer);
  if p > 0 then exit;

  if length(Buffer) = 0 then
    AddDigitToBuffer(0);
  Buffer := Buffer + DefaultFormatSettings.DecimalSeparator;
end;

procedure TNumpad.AddDigitToBuffer(aDigit: TDigit);
begin
  if (aDigit <> 0) and (Buffer = '0') then
    Clear;
  if not((aDigit = 0) and (Buffer = '0')) then
    Buffer := Buffer + IntToStr(aDigit);
end;

function TNumpad.GetClearedBuffer: String;
begin
  Result := '';
end;

constructor TNumpad.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  with ChildSizing do
  begin
    ControlsPerLIne := 3;
    EnlargeHorizontal := crsHomogenousChildResize;
    EnlargeVertical := crsHomogenousChildResize;
    Layout := cclLeftToRightThenTopToBottom;
    ShrinkHorizontal := crsHomogenousChildResize;
    ShrinkVertical := crsHomogenousChildResize;
  end;

  CreateSubControls;
  UndoIconSize := uis24;

  Clear;
end;

procedure TNumpad.Clear;
begin
  Buffer := GetClearedBuffer;
  if Assigned(fOnClear) then
    fOnClear(Self);
end;

procedure TNumpad.InvalidatePreferredSize;
begin
  inherited InvalidatePreferredSize;
  ChildSizing.LeftRightSpacing := BorderSpacing.InnerBorder;
  ChildSizing.TopBottomSpacing := BorderSpacing.InnerBorder;
end;

initialization
  InitUndoImages;
finalization
  FinalizeUndoImages;
end.
