{ Includes touch usable Currency Input Control

  @section Requires cwstrings on *nix

  This unit requires unit @link(cwstrings) on *nix targets for proper conversion
  of system defined Currency String.

  @section Locales on *nix

  This unit uses locales defined in the @link(SysUtils) unit. You may find it
  usefull to use @link(clocale) unit on *nix targets to get these settings from
  the system automatically.

}
unit priceinput;

{$mode objfpc}{$H+}

interface

uses
    Classes
  , numpad
  ;

type

  { Touch usable Currency input control

    This class uses the @link(TNumpad) for creating a touch usable currency
    input control. Thus some behaviour is changed.
    @unorderedList(
      @item(The new property @link(Value) holds the entered value as
        @code(Currency) data type.)
      @link(The property @code(Value) is the leading property (i.e.
        @link(Buffer) is derived from this).)
      @item(The Buffer holds a currency representation of the new property
        @link(Value) according to @link(SysUtils.DefaultFormatSettings).)
      @item(The decimal separator button is changed to add @code('00') to the
        @link(Buffer).)
    )
  }

  TPriceInput = class(TNumpad)
  strict private
    fIsPositive: Boolean;
    fValue: Currency;
    function GetIsPositive: Boolean;
    procedure SetIsPositive(aIsPositive: Boolean);
    procedure SetValue(AValue: Currency);
  strict protected
    function GetClearedBuffer: String; override;
    { Adds @code('00') to buffer

      The button for adding a decimal separator is changed to add @code('00')
      to the buffer.

      @seealso(ChangeDecimalSeparatorToZeroZero)
    }
    procedure AddDecimalSeparatorToBuffer; override;
    { Adds digit as least significant decimal place

      Adds the new digit as the least significant decimal place of the
      @link(Value).
    }
    procedure AddDigitToBuffer(aDigit: TDigit); override;
    { Updates buffer from @link(Value)

      Updates buffer with currency representation of @link(Value)
      @seealso(Value)
      @seealso(Buffer)
    }
    procedure UpdateBuffer; virtual;
    { Changes display for decimal separator button }
    procedure ChangeDecimalSeparatorToZeroZero; virtual;
  public
    constructor Create(TheOwner: TComponent); override;
    procedure Clear; override;
  published
    { The currency value of @link(Buffer)
      @seealso(Buffer)
      @seealso(UpdateBuffer)
    }
    property Value: Currency read fValue write SetValue;
    { Indicates if the current value is positive or negative

      Set this to @code(False) to make the current value negative.
      This will also allow to set the value -0, thus all added digits
      result in a negative value;
    }
    property IsPositive: Boolean read GetIsPositive write SetIsPositive default True;
  end;

procedure Register;

implementation

uses
    math
  , LResources
  , SysUtils
  , StdCtrls
  , LazUTF8
  ;

procedure Register;
begin
  {$I priceinput_icon.lrs}
  RegisterComponents('Touchhelper', [TPriceInput]);
end;

{ TPriceInput }

procedure TPriceInput.SetValue(AValue: Currency);
begin
  fValue := AValue;
  UpdateBuffer;

  FPONotifyObservers(Self, ooChange, nil);
end;

function TPriceInput.GetIsPositive: Boolean;
begin
  Result := fIsPositive;
end;

procedure TPriceInput.SetIsPositive(aIsPositive: Boolean);
begin
  if fIsPositive = aIsPositive then exit;

  fIsPositive:= aIsPositive;
  Value := -Value;
end;

function TPriceInput.GetClearedBuffer: String;
begin
  Result := SysToUTF8(Format('%M', [0.0]));
end;

procedure TPriceInput.AddDecimalSeparatorToBuffer;
begin
  AddDigitToBuffer(0);
  AddDigitToBuffer(0);
end;

procedure TPriceInput.AddDigitToBuffer(aDigit: TDigit);
var
  DecimalFactor: Int64;
  NewAddedValue: Currency;
  OldValueShift: Currency;
  OldExceptMask: TFPUExceptionMask;
  NewValue: Currency;
begin
  OldExceptMask := GetExceptionMask;
  SetExceptionMask([exInvalidOp, exDenormalized, exZeroDivide,
                     exOverflow, exUnderflow, exPrecision]);

  DecimalFactor := trunc(Power(10, DefaultFormatSettings.CurrencyDecimals));
  NewAddedValue := (aDigit / DecimalFactor);
  OldValueShift := (fValue * 10);

  if IsPositive then
    NewValue := OldValueShift + NewAddedValue
  else
    NewValue := OldValueShift - NewAddedValue;
                                  
  SetExceptionMask(OldExceptMask);

  Value := NewValue;
end;

procedure TPriceInput.UpdateBuffer;
var
  NewBuffer: String;
begin
  NewBuffer := SysToUTF8(Format('%M', [fValue]));
  if (not IsPositive) and (fValue = 0) then
    NewBuffer := '-'+NewBuffer;

  Buffer := NewBuffer;
end;

procedure TPriceInput.ChangeDecimalSeparatorToZeroZero;
var
  c: TComponent;
begin
  for c in self do
    if c.Tag = TAG_DECIMALSEPARATOR then
      break;

  if Assigned(c) and (c is TCustomButton) then
    TCustomButton(c).Caption := '00';
end;

constructor TPriceInput.Create(TheOwner: TComponent);
begin
  inherited Create(TheOwner);
  fIsPositive := True;

  ChangeDecimalSeparatorToZeroZero;
  UpdateBuffer;
end;

procedure TPriceInput.Clear;
begin
  inherited Clear;
  fValue := 0;
  fIsPositive := True;
end;

end.

